import optparse
import ConfigParser
import unittest2
import os
import json

try:
  import twentyc.xbahn.xbahn as xbahn
except ImportError, inst:
  print "No xbahn module available"
  xbahn = None

def line():
  return "\n--------------------------------------------------------------------------------"

def runtests(c, tests, extend_parser=None):
  parser = optparse.OptionParser()
  parser.add_option("-c", "--config", dest="configfile", default="$ZENFIRE_HOME/etc/$VODKA_CONFIG", help="path to a vodka config file with couchdb and xbahn information. Defaults to $ZENFIRE_HOME/etc/$VODKA_CONFIG")
  parser.add_option("-v", "--verbose", dest="verbose", action="store_true", help="print details")
  parser.add_option("-t", "--tests", dest="tests", default=None, help="can pass a comma separated list of tests to run only those tests")

  if extend_parser:
    extend_parser(parser)

  (options, args) = parser.parse_args()

  if options.tests:
    tests = options.tests.split(",")

  runner = unittest2.TextTestRunner(verbosity=2)
  suite = unittest2.TestSuite(map(c, tests))

  for test in suite:
    test.options = options

  result = runner.run(suite)
  return result

class TwentyCTestCase(unittest2.TestCase):

  def setUp(self, req_xbahn=False):

    self.configfile = os.path.expandvars(self.options.configfile)
    self.config = ConfigParser.RawConfigParser()
    self.config.read(self.configfile)

    self.verbose = self.options.verbose

    self.info("\nRead config from file: %s" % self.configfile)

    if self.config.has_section("server"):
      srvcnf = dict(self.config.items("server"))
      self.couch_engine = srvcnf.get("couch_engine", "couchdb")
      self.couch_config = dict(self.config.items(self.couch_engine))
    else:
      self.couch_engine = "none"
      self.couch_config = {}
    
    self.databases = {}

    if req_xbahn and self.config.has_section("xbahn"):
      self.xbahn_config = dict(self.config.items("xbahn"))
    elif req_xbahn:
      self.xbahn_config = {}

    if req_xbahn and xbahn:
      self.xbahn = xbahn.xBahnThread(
        self.xbahn_config.get("host"),
        int(self.xbahn_config.get("port",0)),
        self.xbahn_config.get("exchange"),
        None,
        None,
        queue_name="unittest2",
        username=self.xbahn_config.get("username"),
        password=self.xbahn_config.get("password"),
        verbose=self.verbose
      )
      self.xbahn.start()

  def tearDown(self):
    if hasattr(self, "xbahn") and self.xbahn.status == 2:
      self.info("Disconnecting xbahn!")
      self.xbahn.stop()
      self.xbahn.disconnect()

  def info(self, msg):
    if self.verbose:
      print msg

  def require(self, xbahn=False, databases=[]):
    if xbahn and (not hasattr(self, "xbahn") or self.xbahn.status!=2):
      raise self.skipTest("xbahn connection required")
    for db in databases:
      if db not in self.databases.keys():
        raise self.skipTest("database connection '%s'")

  def xbahn_client(self, queue_name):
    return xbahn.xBahnThread(
      self.xbahn_config.get("host"),
      int(self.xbahn_config.get("port")),
      self.xbahn_config.get("exchange"),
      None,
      None,
      queue_name=queue_name,
      username=self.xbahn_config.get("username"),
      password=self.xbahn_config.get("password"),
      verbose=self.verbose
    )
 

  def param_dump(self, param):
    if type(param) == dict:
      return json.dumps(param, indent=2)
    else:
      return "%s" % param

  def add_database(self, name, db):
    self.databases[name] = db

  def add_cli_options(self):
    pass

  def assertIsInstance(self, obj, cl, msg=None):
    self.assertTrue(type(obj) == cl, msg=msg)

  def assertIsIn(self, a, b, msg=None):
    self.assertTrue(a in b, msg=msg)

  def assertIsNotIn(self, a, b, msg=None):
    self.assertTrue(a not in b, msg=msg)
  
  def assertDictEqual(self, a, b ,msg=None, ignore=[]):
    
    msg = "dict mismatch: %s\nA:%s\nB:%s" % (
      msg,
      self.param_dump(a),
      self.param_dump(b)
    )

    for k in ignore:
      if a.has_key(k):
        del a[k]
      if b.has_key(k):
        del b[k]

    self.assertEqual(len(a.keys()), len(b.keys()), msg)
    for k,v in a.items():
      if k in ignore:
        continue

      if type(v) == dict:
        self.assertDictEqual(v, b.get(k), msg, ignore=ignore)
      elif type(v) == list:
        self.assertListEqual(v, b.get(k), msg)
      else:
        m = b.get(k)
        if m != v:
          raise Exception("values not equal: %s" % msg)
  
  def assertListEqual(self, a, b, msg=None):

    n = 0

    msg = "list mismatch: %s\nIS:%s\nSHOULD BE:%s" % (
      msg,
      a,
      b
    )

    self.assertTrue(len(a) == len(b), msg)
      
    for item_a in a:
      item_b = b[n]
      if type(item_a) == dict:
        self.assertDictEqual(item_a, item_b, msg)
      elif type(item_a) == list:
        self.assertListEqual(item_a, item_b, msg)
      else:
        self.assertEqual(item_a, item_b, msg)

      n += 1
    #self.assertTrue(set(a) == set(b), msg)




import os
from Crypto.Cipher import AES
from Crypto.PublicKey import RSA
import Crypto.Util.number

class BillingCrypto:

    @staticmethod
    def generate_key(filename, bits=2048):
        privkey = RSA.generate(bits)
        pubkey = privkey.publickey()
        f = open("%s/keys/%s-priv.pem" % (os.getcwd(), filename),'w')
        f.write(privkey.exportKey('PEM'))
        f.close()

        f = open("keys/%s-pub.pem" % filename,'w')
        f.write(pubkey.exportKey('PEM'))
        f.close()
        
    @staticmethod
    def load_rsa_key(filename, keytype = 'pub', filepath=None):
        
        try:
            if not filepath:
                if (filename[-3:] == "pem"):
                    f = open(filename, 'r')
                else:
                    f = open("%s-%s.pem" % (filename, keytype), 'r')
            else:
                f = open(filepath, "r")
                
            key = RSA.importKey(f.read());
            f.close()
            return key;
        except Exception as e:
            raise e


    @staticmethod
    def iv(length=16):
        return os.urandom(length)
        
    @staticmethod
    def encrypt(pubkey, text, keylength = 24):
        key = ""
        ciphertext = ""
        try:
            key = os.urandom(keylength)
            secret = pubkey.encrypt(key, None)
            iv = BillingCrypto.iv()
            aes_engine = AES.new(key, AES.MODE_CBC, iv)
            
            while(len(text) % 16 != 0):
                text += '\x00'

            ciphertext = aes_engine.encrypt(text)
        except:
            raise        
        return (secret[0].encode('hex'), iv.encode('hex'), ciphertext.encode('hex'))

    @staticmethod
    def decrypt(privkey, key, iv, ciphertext):
        try:
            skey = key.decode('hex')
            txt = ciphertext.decode('hex')
            plainkey = privkey.decrypt(skey)
            civ = iv.decode('hex')
            aes_engine = AES.new(plainkey, AES.MODE_CBC, civ)
            plaintext = aes_engine.decrypt(txt)
            return plaintext
        except:
            raise

if __name__ == "__main__":

    print os.getcwd()
    # run this once to make test keys
    #BillingCrypto.generate_key("test", 2048)

    pubkey = BillingCrypto.load_rsa_key("test", "pub")
    
    txt = "hi this is a test 123"
    (key, iv, cipher) = BillingCrypto.encrypt(pubkey, txt, 16);

    print "Text: %s" % txt
    print "IV: %s" % iv
    print "Key: %s" % key
    print "Cipher: %s" % cipher

    privkey = BillingCrypto.load_rsa_key("test", "priv")
    plaintext = BillingCrypto.decrypt(privkey, key, iv, cipher)
    
    print "Recovered Plaintext: %s" % plaintext


    



import re 

#############################################################################

def perms_structure(perms):
  perms_wc = {}
  for ns, p in perms.items():
    pieces = ns.split(".")
    a = prev = perms_wc 
    n = 0 
    l = len(pieces)
    for k in pieces:
      if n < l-1:
        if not a.has_key(k):
          a[k] = {}
        elif type(a[k]) != dict:
          a["@%s"%k] = a[k]
          a[k] = {}
        a = a[k]
      else:
        if not a.has_key(k):
          a[k] = p
        else:
          a["@%s"%k] = p
      n += 1
   

  return perms_wc 

#############################################################################

def perms_check_fast(perms_structure,namespace):
  pieces = namespace.split(".")
  a = perms_structure
  for k in pieces:
    if not a.has_key(k) and not a.has_key("*"):
      return 0
    a = a.get(k, a.get("*"))
  return a

#############################################################################

def perms_check(perms, prefix, ambiguous=False):
  
  """
  Return the user's perms for the specified prefix

  perms <dict> permissions dict
  prefix <string> namespace to check for perms

  ambiguous <bool=False> if True reverse wildcard matching is active and a perm check for a.b.* will
  be matched by the user having perms to a.b.c or a.b.d - only use this if you know what 
  you are doing.
  """

  try:
   
    token = prefix.split(".")

    i = 1
    l = len(token)
    r = 0
    
    # collect permission rules with a wildcard in them, so we dont do unecessary
    # regex searches later on
    perms_wc = {}
    for ns, p in perms.items():
      if ns.find("*") > -1:
        perms_wc[re.escape(ns).replace("\*", "[^\.]+")] = p

    while i <= l:
      k = ".".join(token[:i])
      matched = False

      # check for exact match
      if perms.has_key(k):
        r = perms.get(k)

      # check for wildcard matches (if any wildcard rules exist)
      elif perms_wc:
        for ns, p in perms_wc.items():
          a = "^%s$" % ns
          b = "^%s\." % ns
          j = len(a)
          u = len(b)
          if j > matched and re.match(a, k):
            r = p
            matched = j
          elif u > matched and re.match(b, k):
            r = p
            matched = u
      
      # if not matched at all and ambiguous flag is true, do ambiguous matching
      if not matched and ambiguous:
        m = "^%s" % re.escape(k).replace("\*", "[^\.]+")
        for ns, p in perms.items():
          if re.match(m, ns) and p > r:
            r = p
            break

      i += 1

    return r
  except:
    raise






ACCESS_READ = 0x01
ACCESS_WRITE = 0x02
ACCESS_XBAHN_WRITE = 0x04

def access_xl(level):
  a = ''
  if level & ACCESS_READ:
    a = 'R'
  if level & ACCESS_WRITE:
    a += 'W'
  if level & ACCESS_XBAHN_WRITE:
    a += 'X'

  if not a:
    a = 'D'

  return a




from ConfigParser import ConfigParser
def dict_conf(filename):
  """
  Return dict object for *.conf file
  """
  config = ConfigParser()
  config.read(filename)
  rv = {}
  for section in config.sections():
    rv[section] = {}
    for key,value in config.items(section):
      rv[section][key] = value.strip('"').strip("'").decode("string_escape")
  return rv



from threading import Thread

###############################################################################
# Classes
###############################################################################

class RunInThread(Thread):
  
  """
  Extends threading.Thread
  
  Example:

  t = RunInThread(my_func)
  t.start("some text")
  will call my_func("some text") in it's own thread
  """

  error_handler = None
  
  def __init__(self, callback):
    
    """
    Init and set callback function, the callback function
    will be executed on run()
    """
    
    Thread.__init__(self)
    self.callback = callback
    self.result_handler = None

  def run(self):
    try:
      r = self.callback(*self.runArgs, **self.runKwargs)
      if self.result_handler:
        self.result_handler(r)
    except Exception, inst:
      if self.error_handler:
        self.error_handler(inst)
      else:
        raise
    del self.callback
    del self.runArgs
    del self.runKwargs

  def start(self, *args, **kwargs):
    
    """
    Set the arguments for the callback function and start the
    thread
    """

    self.runArgs = args
    self.runKwargs = kwargs
    Thread.start(self)





from logging.handlers import SysLogHandler
import socket
try:
	import codecs
except ImportError:
	codecs = None

class UTFFixedSysLogHandler(SysLogHandler):
	"""
	A bug-fix sub-class of SysLogHandler that fixes the UTF-8 BOM syslog
	bug that caused UTF syslog entries to not go to the correct
	facility.  This is fixed by over-riding the 'emit' definition
	with one that puts the BOM in the right place (after prio, instead
	of before it).
	
	Based on Python 2.7 version of logging.handlers.SysLogHandler.
	
	Bug Reference: http://bugs.python.org/issue7077
	"""
	
	def emit(self, record):
		"""
		Emit a record.
		
		The record is formatted, and then sent to the syslog server.  If
		exception information is present, it is NOT sent to the server.
		"""
		msg = self.format(record) + '\000'
		"""
		We need to convert record level to lowercase, maybe this will
		change in the future.
		"""
		prio = '<%d>' % self.encodePriority(self.facility,
					self.mapPriority(record.levelname))
		prio = prio.encode('utf-8')
		# Message is a string. Convert to bytes as required by RFC 5424.
		msg = msg.encode('utf-8')
		if codecs:
			msg = codecs.BOM_UTF8 + msg
		msg = prio + msg
		try:
			if self.unixsocket:
				try:
					self.socket.send(msg)
				except socket.error:
					self._connect_unixsocket(self.address)
					self.socket.send(msg)
			elif self.socktype == socket.SOCK_DGRAM:
				self.socket.sendto(msg, self.address)
			else:
				self.socket.sendall(msg)
		except (KeyboardInterrupt, SystemExit):
			raise
		except:
			self.handleError(record)




from setuptools import setup

setup(
    name='twentyc.tools',
    version='0.1.2',
    author='Twentieth Century',
    author_email='code@20c.com',
    description='various python tool libraries / helpers',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    url='https://github.com/20c/twentyc.tools',
    download_url='https://github.com/20c/twentyc.tools/tarball/0.1.2',
    packages=['twentyc.tools'],
    install_requires=open("requirements.txt").read().split("\n"),
    namespace_packages=['twentyc'],
    zip_safe=False
)



import subprocess
import cmd
import os
import shutil
import subprocess
import commands
import re
import optparse
import ConfigParser
import signal
import time

from twentyc.tools.config import dict_conf

DEBUG = False

class ConfigNotFound(Exception):
  def __init__(self, fileName):
    Exception.__init__(self, "Config not found: %s" % fileName)

###############################################################################

class ConfigValueInvalid(Exception):
  def __init__(self, section, key, expecting):
    Exception.__init__(self, "Invalid value for %s:%s -> expecting %s" % (section,
      key, expecting))

class ConfigValueMissing(Exception):
  def __init__(self, section, key):
    Exception.__init__(self, "Missing value for %s:%s" % (section, key)) 

class InvalidCommand(Exception):
  def __init__(self, name):
    Exception.__init__(self, "Invalid command: %s" % name)

################################################################################

class CLIEnv(cmd.Cmd):

  def __init__(self, name="cli"):
    cmd.Cmd.__init__(self)
    self.serving = False
    self.name = name
    self.xbahn = {}
    self.optparse = optparse.OptionParser()
    self.custom_options()
    self.config_errors = []

  #############################################################################

  def custom_options(self):
    self.optparse.add_option("-c", "--config", dest="configfile", default="%s.conf" % self.name, help="path to cli config file, defaults to ./%s.conf" % self.name)
    
  #############################################################################

  def check_config(self, section, name, has_default=False, value=None, allowEmpty=False):
    
    """
    Check if a config value is set

    Returns

    True - config value is set and valid
    False - config value is not set or invalid
    """

    cfg = self.config
    if not cfg.has_key(section) or not cfg.get(section).has_key(name):
      if not has_default:
        self.config_errors.append("%s: %s -> missing" % (section, name))
        return False
      else:
        return True

    v = cfg.get(section).get(name)

    if v == "" and not allowEmpty:
      self.config_errors.append("%s: %s -> empty" % (section, name))
      return False
    return True

  #############################################################################


  def require_config(self, section, name, has_default=False, value=None, allowEmpty=False):
    cfg = self.config
    if not cfg.has_key(section) or not cfg.get(section).has_key(name):
      if not has_default:
        raise ConfigValueMissing(section, name)
      else:
        return value

    v = cfg.get(section).get(name)

    if v == "" and not allowEmpty:
      raise ConfigValueMissing(section, name)
    return v

  def need_config_file(self):
    return True

  def run(self):

    # parse options
    (options, args) = self.optparse.parse_args()
    self.options = options
    self.args = args

    if len(args) > 0:
      self.run_command = args[0]
    else:
      self.run_command = None

    # load config
    self.configfile = os.path.expandvars(options.configfile)
    
    if self.need_config_file():
      if not os.path.exists(self.configfile):
        raise ConfigNotFound(self.configfile)

    if os.path.exists(self.configfile):
      print "Using config: %s" % self.configfile
      self.config = dict_conf(self.configfile) 
    else:
      self.config = {}

    # shutdown cleanly
    signal.signal(signal.SIGINT, self.shutdown)
    signal.signal(signal.SIGTERM, self.shutdown)

  def connect_xbahn(self):
    import twentyc.xbahn.pool as xbahn_pool
    xbahn_pool.connect_from_config(self.config, start=True)
    self.xbahn = xbahn_pool.instances

  def xbahn_request_to_all(self, namespace, data, **kwargs):
    import twentyc.xbahn.pool as xbahn_pool
    return xbahn_pool.request_to_all(namespace, data, **kwargs)

  def xbahn_find_instance(self, conn_str):
    import twentyc.xbahn.pool as xbahn_pool
    return xbahn_pool.find(conn_str)

  def print_table(self, headers, data):
    try:
      import prettytable
    except Exception, inst:
      print "PrettyTable python module not installed, but is required for this operation"
      return
    print "Printing table with headers: %s" % headers
    table = prettytable.PrettyTable(headers)
    for row in data:
      table.add_row([row.get(k) for k in headers])
    print table

  def serve(self):
    self.serving = True
    while(self.serving):
      self.work()
      time.sleep(1)

  def works(self):
    pass

  def shutdown(self, *args):
    print "Shutting down..."
    self.serving = False
    if self.xbahn:
      import twentyc.xbahn.pool as xbahn_pool
      xbahn_pool.disconnect()

  def require_env_var(self, name, default):
    if not os.environ.has_key(name):
      self.set_env_var(name, default)

  def set_env_var(self, name, value):
    os.environ[name] = str(value)
    print "export %s=%s" % (name, value)

  def copyfiles(self, src, dst, symlinks=False):
    if DEBUG:
      print "Copying files from '%s' to '%s'" % (src, dst)

    if not os.path.exists(dst):
      os.makedirs(dst)

    if not os.path.isdir(src):
      shutil.copy(src, dst)
      return

    for item in os.listdir(src):
      s = os.path.join(src, item)
      d = os.path.join(dst, item)
      if os.path.isdir(s):
        shutil.copytree(s, d, symlinks=symlinks)
      else:
        shutil.copy2(s, d)
  
  def run_shell(self, cmd):
    if DEBUG:
      print "Executing: %s" % cmd
    subprocess.check_call(cmd.split(' '))

  def warn(self, msg):
    print "WARNING: %s"% msg

  def notify(self, msg):
    print msg



################################################################################

def rm_rf(dir):
  if not os.path.exists(dir):
    return
  for root, dirs, files in os.walk(dir, topdown=False):
    for name in files:
      os.remove(os.path.join(root, name))
    for name in dirs:
      os.rmdir(os.path.join(root, name))
  os.rmdir(dir)


###############################################################################

class Command:
  def __init__(self, opt):
    self.options = opt
    self.min_prefix = ""
    self.min_keep_original = True

  def _exec(self, cmd):
    subprocess.check_call(cmd.split(' '))

  def _yuizip(self, src, dst):
    if self.options.nocompress:
      return

    rv, jar = commands.getstatusoutput("ls " + self.options.yuidir + "/build/yuicompressor-*.jar")

    if rv:
      raise RuntimeError("yuicompressor jar not found (" + self.options.yuidir + ")")

    cmd = "java -jar " + jar
    try :
      self._exec(cmd + " " + src + " -o " + dst)
    except OSError :
      print 'encountering OSError while calling java to run yuicompressor, you likely do not have java installed.'
      raise

  def _googlezip(self, src, dst):
    if self.options.nocompress:
      return

    rv, jar = commands.getstatusoutput("ls " + self.options.closurecompilerdir + "/compiler.jar")

    if rv:
      raise RuntimeError("compiler.jar not found (" + self.options.closurecompilerdir + ")")

    cmd = "java -jar " + jar
    try :
      self._exec(cmd + " --js " + src + " --js_output_file " + dst)
    except OSError :
      print 'encountering OSError while calling java to run closure compiler, you likely do not have java installed.'
      raise

  ##############################################################################

  def _install_dir(self, src, targ, DESTDIR):
    dst = DESTDIR + targ
    if not os.path.exists(dst):
      raise RuntimeError("Destination (" + dst + ") does not exist while installing " + src)

    for root, dirs, files in os.walk(src): #, topdown=False):
      if ".svn" in dirs:
        dirs.remove(".svn")
      if ".git" in dirs:
        dirs.remove(".git")

      dir = dst + "/" + root
      print "mkdir " + dir
      if not os.path.exists(dir):
        os.makedirs(dir)

      for name in dirs:
        dir = dst + "/" + root + "/" + name
        print "mkdir " + dir
        if not os.path.exists(dir):
          os.makedirs(dir)
        k = root+"/"+name
        if os.path.islink(k):  
          print "copying symlink dir %s" % k
          shutil.rmtree(dir)
          shutil.copytree(k, dir)

      for name in files:
        srcfile = root + "/" + name
        dstfile_min = dst + "/" + ("%s/%s%s" % (root, self.min_prefix, name))
        dstfile = dst + "/" + srcfile

        if re.match("^%s" %self.min_prefix, name):
          continue

        if name.endswith(".js") and not self.options.nocompress and not name in ["yui.js"]:
          if self.min_keep_original:
            shutil.copyfile(srcfile, dstfile)
          if not self.options.useclosurecompiler:
            self._yuizip(srcfile, dstfile_min)
          else:
            self._googlezip(srcfile, dstfile_min)
        elif name.endswith(".css") and not self.options.nocompress:
          self._yuizip(srcfile, dstfile_min)
        else:
          shutil.copyfile(srcfile, dstfile)
        print "  " + srcfile

  ##############################################################################

  def _install_file(self, src, targ, DESTDIR):
    pass


